#include <joyos.h>
#include <math.h>


#define VPS_TO_FEET (443.4)
#define ROLLER_PORT (1) //red faces up

//Midpoints of each sector
const int midpoints[6][2] = {{1245, 0}, {622, 1078}, {-622, 1078}, {-1245, 0}, {-622, -1078},{622, -1078}};

const int parLever[6][2] = {{1336, -204}, {855, 1055}, {-490, 1226}, {-1304, 185}, {-840, -1069}, {465, -1257}}; //parallel points for levers

const int parRoller[6][2] = {{1413, 259}, {455, 1306}, {-918, 1075}, {-1413, -259}, {-515, -1306}, {918, -1075}}; //parallel points for rollers

const int angle[6][3] = {{-30, -150, 0}, {30, -90, 60}, {90, -30, 120}, {150, 30, 180}, {-150, 90, -120}, {-90, 150, -60}}; //angle headings needed to align robot; 0 = levers; 1 = rollers; 2 = miasmador

const int dropOff[6][2] = {{642,0}, {331, 527}, {-322, 530}, {-647,0}, {-300, -560}, {331,-576}};

const int lever[6][2] = {{1470, -283}, {974, 1107}, {-500, 1380}, {-1500,300}, {-962, -1150}, {512, -1414}};

const int roller[6][2] = {{1521,300}, {515, 1749}, {-1018,1175}, {-1580,-260}, {-515, -1749}, {1018,-1175}};


typedef struct RobotPosition
{
    int x;
    int y;
    int theta;
}RobotPosition;

typedef struct GoalPosition
{
    int x;
    int y;
}GoalPosition;


//Headers from PathDrive.c
//Both values currently in inches
#define ROBOT_HEIGHT 10
#define CAMERA_HEIGHT 129

//Function declarations

RobotPosition getCurrentPosition();
int getSimpleHeading(int, int, int, int);
float getSimpleDistanceTo(int, int, int, int);
void recalibrateGyro();
int angleModulus(int);
void proportionalDrive (int, int, int);
void goTowards(int, int, int);
float getSimpleDistanceTo(int, int, int ,int);
void turn(int);
void stop();
void rollerBlue();
void rollerRed();
void positionLever();
void pressLever();
void holdBalls();
void releaseBalls();
void driveStraight(int, int, int);
RobotPosition findOpponent();
void shake();
void driveStraight();
bool goToPointInternal(int, int, int, int);
void backOut();
void shake();

//End declarations

int currentTerritory(int x, int y) //input: coordinates of point
{
    int theta = atan2(y, x)* 180/3.14159265;
    if (theta>0)
    {
        for (int i=0; i<4; i++)
        {
            if((theta>((60*i)-30)) && (theta<((60*i)+30)))
            {
                return i;
            }
        }
    }
    else
    {
        if ((theta<0) && (theta>-30))
            return 0;
        else if ((theta<-30) && (theta>-90))
            return 5;
        else if ((theta<-90) && (theta>-150))
            return 4;
        else
            return 3;
    }
}

void goToPoint(int x, int y, int direction, int tolerance)
{
    int attempts = 0;
    RobotPosition me = getCurrentPosition();
    int myTerr = currentTerritory(me.x, me.y);
    int orig_x = x;
    int orig_y = y;
    while (!goToPointInternal(x, y, direction, tolerance))
    {
        backOut(x,y);
        attempts++;
        me = getCurrentPosition();
        myTerr = currentTerritory(me.x, me.y);
        if (attempts == 1)
            goToPointInternal(midpoints[myTerr][1], midpoints[myTerr][1], direction, tolerance);
        else if (attempts == 2)
        {
            x = midpoints[(myTerr+1)%6][0];
            y = midpoints[(myTerr+1)%6][1];
        }
        else if (attempts == 3)
        {
            int destTerr = myTerr - 1;
            if (destTerr<0)
            {
                destTerr +=6;
            }
            x = midpoints[destTerr][0];
            y = midpoints[destTerr][1];
        }
        else
        {
            shake();
            x = orig_x;
            y = orig_y;
            attempts = 0;
        }
    }
}

bool getColor() //only call in usetup when robot is in starting position
{
    RobotPosition me = getCurrentPosition();
    if (me.x>0)
        return true;
    else
        return false;
}


void circleBlue() //in case we hit opponent and get past 10 seconds, the robot will go to regular game play
{
    int32_t circleStartTime = get_time();
    turn(90);
    goToPoint(-727, 1300, 0, VPS_TO_FEET/2);
    if ((get_time() - circleStartTime) > 10000)
        return;
    goToPoint(-1100, -1050, 0, VPS_TO_FEET/2);
    if ((get_time() - circleStartTime) > 10000)
        return;
    goToPoint(midpoints[5][0], midpoints[5][1], 0, VPS_TO_FEET/2);
}

void circleRed()
{
    int32_t circleStartTime = get_time();
    turn(-90);
    goToPoint(680, -1400, 0, VPS_TO_FEET/2);
    if ((get_time() - circleStartTime) > 10000)
        return;
    goToPoint(775, 1156, 0, VPS_TO_FEET/2);
    if ((get_time() - circleStartTime) > 10000)
        return;
    goToPoint(midpoints[2][0], midpoints[2][1], 0, VPS_TO_FEET/2);
}

void goToTerritory (int goalTerr) //goes to midpoint
{
    RobotPosition me = getCurrentPosition();
    int myTerr = currentTerritory(me.x, me.y);
    int diff = goalTerr - myTerr;

	//Print out for Testing
	//printf("Difference: %d \n", diff);

    if (abs(diff) < 2 || abs(diff) == 5)
    {
        goToPoint((midpoints[goalTerr][0]), (midpoints[goalTerr][1]),0, VPS_TO_FEET / 4);
    }
    else if (abs(diff) == 2)
    {
        if (diff>0)
        {
            for (int i=(currentTerritory(me.x, me.y)+1); i<(currentTerritory(me.x, me.y)+3); i++)
            {
                goToPoint((midpoints[i][0]), (midpoints[i][1]),0, VPS_TO_FEET / 4);
            }
        }
        else
        {
            for (int j=(currentTerritory(me.x, me.y)-1); j>(currentTerritory(me.x, me.y)-3); j--)
            {
                goToPoint((midpoints[j][0]), (midpoints[j][1]),0, VPS_TO_FEET / 4);
            }
        }
    }
    else if (abs(diff) == 4)
    {
        if (currentTerritory(me.x, me.y) == 1)
        {
            goToTerritory(0);
            goToTerritory(5);
        }
        else if (currentTerritory(me.x, me.y) == 0)
        {
            goToTerritory(5);
            goToTerritory(4);
        }
        else if (currentTerritory(me.x, me.y) == 4)
        {
            goToTerritory(5);
            goToTerritory(0);
        }
        else //currentTerritory == 5
        {
            goToTerritory(0);
            goToTerritory(1);
        }
    }
    else //3 or -3
    {
        if (diff>0)
        {
            for (int k=(currentTerritory(me.x, me.y)+1); k<(currentTerritory(me.x, me.y)+4); k++)
            {
                goToPoint((midpoints[k][0]), (midpoints[k][1]),0, VPS_TO_FEET / 4);
            }
        }
        else
        {
            for (int k=(currentTerritory(me.x, me.y)-1); k>(currentTerritory(me.x, me.y)-4); k--)
            {
                goToPoint((midpoints[k][0]), (midpoints[k][1]),0, VPS_TO_FEET / 4);

            }
        }
    }
}

void turn(int heading)
//Turns until we're facing a given direction
//Includes timeout
{
    int32_t startTime = get_time();
    while (angleModulus(heading - angleModulus(gyro_get_degrees())) > 5)
    {
        goTowards(0, heading, 1);
        //Timeout
        if (get_time() - startTime > 5000)
        {
            //Save our current position
            RobotPosition saved = getCurrentPosition();
            //Go to midpoint
            int terr = currentTerritory(saved.x, saved.y);
            goToPoint(midpoints[terr][0], midpoints[terr][1], 0, VPS_TO_FEET/4);
            //Back to saved position
            goToPoint(saved.x, saved.y, 0, VPS_TO_FEET/4);
            //Reset timeout
            startTime = get_time();
        }
    }
}

//before calling goToRoller, goToLever, or goToMiasmador, make sure you're already in that territory!
void goToRoller(bool teamBlue) //input bool teamblue -> if true = 1; false = 0
{
    RobotPosition me = getCurrentPosition();
    RobotPosition you = findOpponent();
    int terr = currentTerritory(me.x, me.y);
    int attempts = 0;
    while ((game.territories[terr].owner != 14)
           && (attempts<2)
           && (getSimpleDistanceTo(roller[terr][0], roller[terr][1], you.x, you.y) > VPS_TO_FEET))
    {
        //Go to midpoint and turn to face correct direction
        goToPoint(midpoints[terr][0], midpoints[terr][1], 0, VPS_TO_FEET/4);
        me = getCurrentPosition();
        terr = currentTerritory(me.x, me.y);
        turn(angle[terr][1]);

        //Turn on rollers in the right direction
        if(teamBlue)
		{
			rollerBlue();
		}
		else
        {
			rollerRed();
		}
        //Back into roller
        int32_t startTime = get_time();
        while ((get_time() - startTime < 4000)
               && game.territories[terr].owner !=14)
        {
            proportionalDrive(roller[terr][0], roller[terr][1], -1);
            copy_objects();
        }
        stop();
        motor_set_vel(ROLLER_PORT, 0);
        you = findOpponent();
        if (game.territories[terr].owner !=14)
        {
            attempts++;
        }
    }
    goToPoint(midpoints[terr][0], midpoints[terr][1], 0, VPS_TO_FEET/4);
    if (attempts == 2)
    {
        int retry = (terr+2)%6;
        you = findOpponent();
        int yourTerr = currentTerritory(you.x, you.y);
        if ((retry == yourTerr) || (yourTerr == ((terr+1)%6)))
        {
            retry = terr-2;
            if (retry<0)
            {
                retry+=6;
            }
        }
        goToTerritory(retry);
    }
}

bool goToLever()
{
    bool output = false;
    RobotPosition me = getCurrentPosition();
    RobotPosition you = findOpponent();
    int terr = currentTerritory(me.x, me.y);
    int attempts = 0;
    //Go until we have all the balls we can claim
    copy_objects();
    while ((game.territories[terr].remaining > 0)
           && (attempts<3)
           && game.territories[terr].owner == 14
           && (getSimpleDistanceTo(lever[terr][0], lever[terr][1], you.x, you.y)) > VPS_TO_FEET)
    {
        recalibrateGyro();
        //Drive to parallel point and turn
        goToPoint(parLever[terr][0], parLever[terr][1], 0, VPS_TO_FEET / 6);
        me = getCurrentPosition();
        terr = currentTerritory(me.x, me.y);
        turn(angle[terr][0]);
        //Lever up
        positionLever();

        //Drive into lever, going straight forward
        driveStraight(lever[terr][0], lever[terr][1], (int)(VPS_TO_FEET/4));
        stop();

        int startBalls = game.territories[terr].remaining;
        pressLever();
        recalibrateGyro();
        copy_objects();
        //Did we actually get a ball?
        if (game.territories[terr].remaining != startBalls)
        {
            //Get the rest of the balls.  If we missed, go back and
            //reposition to try again.
            for (int i = 0; i<game.territories[terr].remaining; i++)
            {pressLever();}
            output = true;
        }
        pause (500);

        copy_objects();
        attempts++;
        you = findOpponent();
    }
    //back away from lever
    goToPoint(midpoints[terr][0], midpoints[terr][1], -1, VPS_TO_FEET/4);
    if (output == false)
    {
        int retry = (terr+2)%6;
        you = findOpponent();
        int yourTerr = currentTerritory(you.x, you.y);
        if ((retry == yourTerr) || (yourTerr == ((terr+1)%6)))
        {
            retry = terr-2;
            if (retry<0)
            {
                retry+=6;
            }
        }
        goToTerritory(retry);
    }
    return output;
}

void goToMiasmadorInternal() // call only when at the midpoint of the right territory!
{
    RobotPosition me = getCurrentPosition();
    int myTerr = currentTerritory(me.x, me.y);
    turn (angle[myTerr][2]);
    //move to miasmador
    goToPoint(dropOff[myTerr][0], dropOff[myTerr][1], -1, VPS_TO_FEET/12);
    me = getCurrentPosition();
    myTerr = currentTerritory(me.x, me.y);
	//Turn to face miasmador, again, just in case.
	turn(angle[myTerr][2]);
    stop();
	//release balls
	releaseBalls();
	pause(1000); //give enough time for balls to fall out
	shake();
	stop();
	//Put down holding mechanism
	holdBalls();
    goToPoint(midpoints[myTerr][0], midpoints[myTerr][1], 0, VPS_TO_FEET/4);
    turn(angleModulus(90 + myTerr * 60));
}

int findMiasmador(bool teamBlue, int attempts, int previousTry)
{
    RobotPosition me = getCurrentPosition();
    int myTerr = currentTerritory(me.x, me.y);
    int goalTerr = 7;
    if(teamBlue && (myTerr==1 || myTerr==2 || myTerr==3))
    {
        return myTerr;
    }
    else if (teamBlue && (myTerr!=1 || myTerr!=2 || myTerr!=3))
    {
        if (attempts == 0)
        {
            if((myTerr == 0) || (myTerr == 5))
            {
                goalTerr = 1;
            }
            else if(myTerr == 4)
            {
                goalTerr = 3;
            }
        }
        else if (attempts >= 1)
        {
            goToTerritory(5);
            if(previousTry == 1)
            {
                goalTerr = 3;
            }
            else if(previousTry == 3)
            {
                goalTerr = 1;
            }
        }

    }
    else if (!teamBlue && (myTerr==0 || myTerr==4 || myTerr==5))
    {
        return myTerr;
    }

    else if (!teamBlue && (myTerr!=0 || myTerr!=4 || myTerr!=5))
    {
        if (attempts == 0)
        {
            if((myTerr == 1) || (myTerr == 2))
            {
                goalTerr = 0;
            }
            else if(myTerr == 3)
            {
                goalTerr = 4;
            }
        }
        else if (attempts >= 1)
        {
            goToTerritory(2);
            if(previousTry == 0)
            {
                goalTerr = 4;
            }
            else if(previousTry == 4)
            {
                goalTerr = 0;
            }
        }
    }
    return goalTerr;
}

void goToMiasmador(bool teamBlue)
{
    int MiasmadorAttempts = 0;
    int previousTry = 7;
    RobotPosition me = getCurrentPosition();
	int terr = currentTerritory(me.x, me.y);

	//First, go to the current territory's midpoint
	goToTerritory(terr);

    while(MiasmadorAttempts < 3)
    {
        //Then go to a good territory

        previousTry = findMiasmador(teamBlue, MiasmadorAttempts, previousTry);
        goToTerritory(previousTry);

        me = getCurrentPosition();
        terr = currentTerritory(me.x, me.y); //check that we ended up in a good territory
        if(teamBlue && (terr==1 || terr==2 || terr==3)) // in a good territory (not opponent side)
        {
            goToMiasmadorInternal();
            return;
        }
        else if(!teamBlue && (terr==0 || terr==4 || terr==5))
        {
            goToMiasmadorInternal();
            return;
        }
        else
        {
            MiasmadorAttempts++;
        }
    }
}












